import torch
from aftab.maps import networks_map
from .AftabBaseMixin import AftabBaseMixin

_DISTRIBUTIONAL_NETWORKS = {
    "distributional",
    "distributional-duelling",
    "distributional-bootstrapped-duelling",
}
_BOOTSTRAPPED_NETWORKS = {
    "bootstrapped",
    "bootstrapped-duelling",
    "distributional-bootstrapped-duelling",
}


class AftabNetworkMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    def __channels_last_enabled(self) -> bool:
        return bool(getattr(self, "channels_last"))

    def __dummy_input(self) -> torch.Tensor:
        if not hasattr(self, "frame_stack"):
            raise AttributeError("Expected `frame_stack` to be defined.")
        if not hasattr(self, "device"):
            raise AttributeError("Expected `device` to be defined.")

        sample = torch.randn(
            1,
            self.frame_stack,
            84,
            84,
            device=self.device,
        )
        if self.__channels_last_enabled():
            return sample.contiguous(memory_format=torch.channels_last)
        return sample

    def __distributional_sigma(self) -> float:
        sigma = getattr(self, "distributional_sigma", None)
        if sigma is None:
            bins = int(getattr(self, "distributional_bins"))
            min_value = float(getattr(self, "distributional_min_value"))
            max_value = float(getattr(self, "distributional_max_value"))
            if bins <= 0:
                raise ValueError("Expected `distributional_bins` to be positive.")
            if max_value <= min_value:
                raise ValueError(
                    "Expected `distributional_max_value` to be greater than "
                    "`distributional_min_value`."
                )
            bin_width = (max_value - min_value) / bins
            distributional_sigma_ratio = float(
                getattr(self, "distributional_sigma_ratio")
            )
            sigma = bin_width * distributional_sigma_ratio
            sigma_name = "`distributional_sigma_ratio`"
        else:
            sigma = float(sigma)
            sigma_name = "`distributional_sigma`"

        if sigma <= 0.0:
            raise ValueError(f"Expected {sigma_name} to be positive.")
        return sigma

    def __network_kwargs(self) -> dict:
        kwargs = {}
        if self.network in _DISTRIBUTIONAL_NETWORKS:
            kwargs.update(
                distributional_bins=int(getattr(self, "distributional_bins")),
                distributional_min_value=float(
                    getattr(self, "distributional_min_value")
                ),
                distributional_max_value=float(
                    getattr(self, "distributional_max_value")
                ),
                distributional_sigma=self.__distributional_sigma(),
            )

        if self.network not in _BOOTSTRAPPED_NETWORKS:
            return kwargs

        bootstrap_heads = int(getattr(self, "bootstrap_heads"))
        if bootstrap_heads <= 0:
            raise ValueError("Expected `bootstrap_heads` to be positive.")
        kwargs["bootstrap_heads"] = bootstrap_heads

        return kwargs

    def __handle_channel_last(self):
        if self.__channels_last_enabled():
            self._network.to(device=self.device, memory_format=torch.channels_last)
        else:
            self._network.to(self.device)

    @torch.inference_mode()
    def __handle_dummy_pass(self):
        self._network(self.__dummy_input())

    def __handle_compilation(self):
        if not hasattr(torch, "compile"):
            return

        self._network = torch.compile(self._network)

    def __get_network_instance(self):
        if self.network not in networks_map:
            raise ValueError(
                f"Invalid value for `network`: {self.network!r}. "
                f"Expected one of {tuple(networks_map)}."
            )

        return networks_map[self.network]

    def _initialize_network(self, action_dimension: int):
        network_instance = self.__get_network_instance()
        self.flush_verbose(f"Network: {network_instance.__name__}")
        self._network = network_instance(
            action_dimension=action_dimension,
            embedding_dimension=self.embedding_dimension,
            encoder=self.encoder,
            channels_last=self.__channels_last_enabled(),
            **self.__network_kwargs(),
        )
        self.__handle_channel_last()
        self.__handle_dummy_pass()
        self.__handle_compilation()



from .BaseNetwork import BaseNetwork
from .PQNNetwork import PQNNetwork
from .DuellingNetwork import DuellingNetwork
from .BootstrappedNetwork import BootstrappedNetwork
from .BootstrappedDuellingNetwork import BootstrappedDuellingNetwork
from .DistributionalNetwork import DistributionalNetwork
from .DistributionalDuellingNetwork import DistributionalDuellingNetwork
from .DistributionalBootstrappedDuellingNetwork import (
    DistributionalBootstrappedDuellingNetwork,
)



import torch


@torch.jit.script
def lambda_returns(
    rewards: torch.Tensor,
    terminations: torch.Tensor,
    next_q: torch.Tensor,
    gamma: float,
    return_lambda: float,
) -> torch.Tensor:
    trajectory_length = rewards.size(0)
    output = torch.zeros_like(rewards)
    not_done_last = 1.0 - terminations[-1]
    accumulated_return = rewards[-1] + gamma * not_done_last * next_q[-1]
    output[-1] = accumulated_return
    for t in range(trajectory_length - 2, -1, -1):
        not_done = 1.0 - terminations[t]
        bootstrap = next_q[t]
        mix = (1.0 - return_lambda) * bootstrap + return_lambda * accumulated_return
        accumulated_return = rewards[t] + gamma * not_done * mix
        output[t] = accumulated_return
    return output



from .AftabBaseMixin import AftabBaseMixin
from .AftabTrainingResultsMixin import AftabTrainingResultsMixin
from .AftabEnvironmentMixin import AftabEnvironmentMixin
from .AftabActionsMixin import AftabActionsMixin
from .AftabNetworkMixin import AftabNetworkMixin
from .AftabQValueMixin import AftabQValueMixin
from .AftabLossMixin import AftabLossMixin
from .AftabReturnsMixin import AftabReturnsMixin
from .AftabTrainMixin import AftabTrainMixin
from .AftabOptimizerMixin import AftabOptimizerMixin
from .AftabStoreModelMixin import AftabStoreModelMixin



import torch
from aftab.constants import ModuleType
from .HadamaxLayerNorm2d import HadamaxLayerNorm2d


class HadamaxBlock(torch.nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        padding: int,
        stride: int,
        pool_kernel: int,
        pool_stride: int,
        pool_padding: int = 0,
        chi: ModuleType = torch.nn.GELU,
        psi: ModuleType = torch.nn.GELU,
    ):
        super().__init__()

        self.convolutional = torch.nn.Conv2d(
            in_channels,
            out_channels * 2,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )
        self.normalization = HadamaxLayerNorm2d(out_channels)
        self.pool = torch.nn.MaxPool2d(
            kernel_size=pool_kernel, stride=pool_stride, padding=pool_padding
        )

        self.same_activations = chi == psi
        self.chi = chi()
        self.psi = psi() if not self.same_activations else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.normalization(self.convolutional(x))
        if self.same_activations:
            a, b = torch.chunk(self.chi(x), 2, dim=1)
        else:
            a, b = torch.chunk(x, 2, dim=1)
            a, b = self.chi(a), self.psi(b)
        return self.pool(a * b)



import jax
import flax.linen as nn


class LayerNorm2d(nn.Module):
    epsilon: float = 1e-6

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        return nn.LayerNorm(
            epsilon=self.epsilon,
            reduction_axes=-1,
            feature_axes=-1,
        )(x)



from .encoders_map import encoders_map
from .networks_map import networks_map
from .optimizers_maps import optimizers_map

from .acceptable_frames_map import acceptable_frames_map

from .encoders_map import encoders_map as encoder_map
from .networks_map import networks_map as network_map
from .optimizers_maps import optimizers_map as optimizer_map



from aftab.encoders import NatureDQNEncoder
from aftab.encoders import AlphaEncoder
from aftab.encoders import BetaEncoder
from aftab.encoders import GammaEncoder
from aftab.encoders import DeltaEncoder
from aftab.encoders import EpsilonEncoder
from aftab.encoders import EtaEncoder
from aftab.encoders import ZetaEncoder
from aftab.encoders import ThetaEncoder
from aftab.encoders import HadamaxNatureDQNEncoder
from aftab.encoders import HadamaxGammaEncoderV1
from aftab.encoders import HadamaxGammaEncoderV2
from aftab.encoders import HadamaxEpsilonEncoder
from aftab.encoders import HadamaxZetaEncoder
from aftab.encoders import HadamaxDeltaEncoder

encoders_map = {
    "nature": NatureDQNEncoder,
    "dqn": NatureDQNEncoder,
    "alpha": AlphaEncoder,
    "beta": BetaEncoder,
    "gamma": GammaEncoder,
    "delta": DeltaEncoder,
    "epsilon": EpsilonEncoder,
    "eta": EtaEncoder,
    "zeta": ZetaEncoder,
    "theta": ThetaEncoder,
    "hadamax": HadamaxNatureDQNEncoder,
    "dqnhadamax": HadamaxNatureDQNEncoder,
    "pqnhadamax": HadamaxNatureDQNEncoder,
    "hadamaxgammav1": HadamaxGammaEncoderV1,
    "gammahadamaxv1": HadamaxGammaEncoderV1,
    "hadamaxgammav2": HadamaxGammaEncoderV2,
    "gammahadamaxv2": HadamaxGammaEncoderV2,
    "hadamaxepsilon": HadamaxEpsilonEncoder,
    "epsilonhadamax": HadamaxEpsilonEncoder,
    "hadamaxzeta": HadamaxZetaEncoder,
    "zetahadamax": HadamaxZetaEncoder,
    "hadamaxdelta": HadamaxDeltaEncoder,
    "deltahadamax": HadamaxDeltaEncoder,
}



import torch
import numpy
import time
from typing import Optional
from .AftabBaseMixin import AftabBaseMixin
from ..common import RolloutBuffer


_TORCH_TO_NUMPY_DTYPE = {
    torch.bool: numpy.bool_,
    torch.uint8: numpy.uint8,
    torch.int8: numpy.int8,
    torch.int16: numpy.int16,
    torch.int32: numpy.int32,
    torch.int64: numpy.int64,
    torch.float16: numpy.float16,
    torch.float32: numpy.float32,
    torch.float64: numpy.float64,
}


class AftabTrainMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    def __autocast_float16_enabled(self) -> bool:
        return (
            bool(getattr(self, "autocast_float16", False))
            and self.device.type != "cpu"
            and torch.amp.autocast_mode.is_autocast_available(self.device.type)
        )

    def __autocast_float16(self):
        return torch.amp.autocast(
            device_type=self.device.type,
            dtype=torch.float16,
            enabled=self.__autocast_float16_enabled(),
        )

    def __numpy_array(self, value, dtype: torch.dtype) -> numpy.ndarray:
        numpy_dtype = _TORCH_TO_NUMPY_DTYPE.get(dtype)
        if isinstance(value, numpy.ndarray):
            array = value
            if numpy_dtype is not None and array.dtype != numpy_dtype:
                array = array.astype(numpy_dtype, copy=False)
        else:
            array = numpy.asarray(value, dtype=numpy_dtype)

        if not array.flags.c_contiguous:
            array = numpy.ascontiguousarray(array)
        return array

    def __tensor(self, value, dtype: torch.dtype) -> torch.Tensor:
        if isinstance(value, torch.Tensor):
            return value.to(dtype=dtype, device=self.device)

        tensor = torch.from_numpy(self.__numpy_array(value, dtype))
        if tensor.dtype != dtype:
            tensor = tensor.to(dtype=dtype)
        if tensor.device == self.device:
            return tensor
        return tensor.to(
            device=self.device,
            non_blocking=self.device.type == "cuda",
        )

    def __copy_array_to_tensor(
        self,
        target: torch.Tensor,
        value,
        dtype: torch.dtype,
    ) -> None:
        if isinstance(value, torch.Tensor):
            source = value.to(dtype=dtype, device=target.device)
        else:
            source = torch.from_numpy(self.__numpy_array(value, dtype))
            if source.dtype != dtype:
                source = source.to(dtype=dtype)
        target.copy_(source, non_blocking=target.device.type == "cuda")

    def __distributional_value_clip_enabled(self) -> bool:
        return bool(getattr(self._network, "distributional", False)) and (
            float(getattr(self, "distributional_value_clip")) > 0.0
        )

    def __bootstrapped_enabled(self) -> bool:
        return bool(getattr(self._network, "bootstrapped", False))

    def __get_bootstrap_heads(self) -> int:
        return int(getattr(self._network, "bootstrap_heads", 1))

    def __sample_bootstrap_heads(self, size: int) -> torch.Tensor:
        return torch.randint(self.__get_bootstrap_heads(), (size,), device=self.device)

    def __sample_bootstrap_masks(self) -> torch.Tensor:
        probability = float(getattr(self, "bootstrap_probability", 1.0))
        if probability <= 0.0 or probability > 1.0:
            raise ValueError("Expected `bootstrap_probability` to be in (0, 1].")

        mask_shape = (self.train_environments, self.__get_bootstrap_heads())
        if probability == 1.0:
            return torch.ones(mask_shape, dtype=torch.float32, device=self.device)
        return (torch.rand(mask_shape, device=self.device) < probability).float()

    def __get_bootstrapped_vote_q_values(self, q_heads: torch.Tensor) -> torch.Tensor:
        head_actions = q_heads.argmax(dim=-1)
        votes = torch.nn.functional.one_hot(
            head_actions,
            num_classes=q_heads.shape[-1],
        )
        return votes.sum(dim=1).to(dtype=q_heads.dtype)

    def __resample_terminated_bootstrap_heads(
        self,
        *,
        active_heads: Optional[torch.Tensor],
        terminations,
    ) -> None:
        if active_heads is None:
            return

        done_indices = numpy.nonzero(terminations)[0]
        done_count = int(done_indices.size)
        if done_count == 0:
            return

        done_indices = torch.from_numpy(done_indices).to(
            device=self.device,
            non_blocking=self.device.type == "cuda",
        )
        active_heads[done_indices] = self.__sample_bootstrap_heads(done_count)

    def __get_step_q_values(
        self,
        *,
        observations: torch.Tensor,
        active_heads: Optional[torch.Tensor],
    ):
        if not self.__bootstrapped_enabled():
            q_values_all = self.get_q_values(
                float_observations=observations,
                gradient=False,
            )
            q_values_train = q_values_all[: self.train_environments]
            q_values_test = q_values_all[self.train_environments :]
            state_q_values = q_values_train.float().max(dim=-1).values
            return q_values_train, q_values_test, state_q_values, None

        if active_heads is None:
            raise RuntimeError("Expected active bootstrapped heads.")

        q_heads_all = self._network.get_q_heads(observations)
        q_heads_train = q_heads_all[: self.train_environments]
        q_values_train = self._network.gather_q_heads(
            q_heads=q_heads_train,
            head_indices=active_heads[: self.train_environments],
        )
        q_values_test = self.__get_bootstrapped_vote_q_values(
            q_heads_all[self.train_environments :]
        )
        state_q_values = q_heads_train.float().max(dim=-1).values
        return q_values_train, q_values_test, state_q_values, q_heads_train

    def __initialize_training(self, environment: str, seed: int):
        train_environment, test_environment, action_dimension, observation_shape = (
            self.make_environments(environment=environment, seed=seed)
        )

        self._initialize_network(action_dimension=action_dimension)
        self._initialize_optimizer()

        scaler = torch.amp.GradScaler(
            enabled=self.device.type == "cuda" and self.__autocast_float16_enabled()
        )

        if test_environment is None:
            (train_observation, _), (test_observation, _) = train_environment.reset_split()
        else:
            train_observation, _ = train_environment.reset()
            test_observation, _ = test_environment.reset()
        observation = torch.empty(
            (self.total_environments, *observation_shape),
            dtype=torch.uint8,
            device=self.device,
        )
        self.__copy_array_to_tensor(
            observation[: self.train_environments],
            train_observation,
            torch.uint8,
        )
        self.__copy_array_to_tensor(
            observation[self.train_environments :],
            test_observation,
            torch.uint8,
        )
        episode_returns = numpy.zeros(self.total_environments, dtype=numpy.float32)
        return (
            train_environment,
            test_environment,
            observation_shape,
            scaler,
            observation,
            episode_returns,
        )

    def __allocate_buffers(self, *, observation_shape: tuple) -> RolloutBuffer:
        return RolloutBuffer(
            observation_shape=observation_shape,
            steps_per_update=self.steps_per_update,
            train_environments=self.train_environments,
            device=self.device,
            bootstrapped=self.__bootstrapped_enabled(),
            bootstrap_heads=self.__get_bootstrap_heads(),
            store_old_q_values=self.__distributional_value_clip_enabled(),
        )

    def __get_score_rewards(
        self,
        *,
        reward_train,
        reward_test,
        info_train,
        info_test,
        output,
    ):
        score_reward_train = info_train.get("reward", None)
        score_reward_test = info_test.get("reward", None)
        if not isinstance(score_reward_train, numpy.ndarray) or not isinstance(
            score_reward_test, numpy.ndarray
        ):
            output[: self.train_environments] = reward_train
            output[self.train_environments :] = reward_test
            return output

        if score_reward_train.ndim == 0:
            return numpy.stack([score_reward_train, score_reward_test]).astype(
                numpy.float32,
                copy=False,
            )
        else:
            output[: self.train_environments] = score_reward_train
            output[self.train_environments :] = score_reward_test
        return output

    def __get_episode_terminations(
        self,
        *,
        info_train,
        info_test,
        fallback_terminations,
        output,
    ):
        episode_done_train = info_train.get("terminated", None)
        episode_done_test = info_test.get("terminated", None)
        if not isinstance(episode_done_train, numpy.ndarray) or not isinstance(
            episode_done_test, numpy.ndarray
        ):
            return fallback_terminations

        if episode_done_train.ndim == 0:
            return fallback_terminations

        output[: self.train_environments] = episode_done_train
        output[self.train_environments :] = episode_done_test
        return output

    def __record_completed_episodes(self, *, episode_returns, terminations):
        idx = numpy.nonzero(terminations)[0]
        if idx.size == 0:
            return

        scores = episode_returns[idx]
        split = idx < self.train_environments
        self.results.rewards.train.extend(scores[split].tolist())
        self.results.rewards.test.extend(scores[~split].tolist())
        episode_returns[idx] = 0

    @torch.inference_mode()
    def __collect_trajectories(
        self,
        *,
        frame_count: int,
        train_environment,
        test_environment,
        episode_returns,
        observation: torch.Tensor,
        rollout_buffer: RolloutBuffer,
        active_heads: Optional[torch.Tensor],
    ):
        actions_train_cpu = torch.empty(self.train_environments, dtype=torch.int64)
        actions_test_cpu = torch.empty(self.test_environments, dtype=torch.int64)
        actions_train = actions_train_cpu.numpy()
        actions_test = actions_test_cpu.numpy()
        terminations = numpy.empty(self.total_environments, dtype=numpy.bool_)
        episode_terminations = numpy.empty(self.total_environments, dtype=numpy.bool_)
        score_rewards_output = numpy.empty(
            self.total_environments,
            dtype=numpy.float32,
        )
        full_bootstrap_masks = None
        if (
            rollout_buffer.bootstrap_masks is not None
            and float(getattr(self, "bootstrap_probability", 1.0)) == 1.0
        ):
            full_bootstrap_masks = torch.ones(
                (self.train_environments, self.__get_bootstrap_heads()),
                dtype=torch.float32,
                device=self.device,
            )

        for step in range(self.steps_per_update):
            train_observation = observation[: self.train_environments]
            epsilon_value = self._network.epsilon.get(
                frame_count,
                self.effective_frames,
            )

            with self.__autocast_float16():
                q_values_train, q_values_test, state_q_values, q_heads_train = (
                    self.__get_step_q_values(
                        observations=observation,
                        active_heads=active_heads,
                    )
                )
            actions_train_tensor, actions_test_tensor = self.get_action_tensors(
                q_values_train=q_values_train,
                q_values_test=q_values_test,
                epsilon_value=epsilon_value,
            )
            actions_train_cpu.copy_(actions_train_tensor)
            actions_test_cpu.copy_(actions_test_tensor)
            old_q_values = None
            if rollout_buffer.old_q_values is not None:
                if q_heads_train is None:
                    old_q_values = q_values_train.gather(
                        1,
                        actions_train_tensor.unsqueeze(1),
                    ).squeeze(1)
                else:
                    action_indices = actions_train_tensor.reshape(-1, 1, 1).expand(
                        -1,
                        self.__get_bootstrap_heads(),
                        1,
                    )
                    old_q_values = q_heads_train.gather(2, action_indices).squeeze(2)

            if test_environment is None:
                train_step, test_step = train_environment.step_split(
                    actions_train,
                    actions_test,
                )
                (
                    next_observation_train,
                    reward_train,
                    termination_train,
                    truncation_train,
                    info_train,
                ) = train_step
                (
                    next_observation_test,
                    reward_test,
                    termination_test,
                    truncation_test,
                    info_test,
                ) = test_step
            else:
                (
                    next_observation_train,
                    reward_train,
                    termination_train,
                    truncation_train,
                    info_train,
                ) = train_environment.step(actions_train)

                (
                    next_observation_test,
                    reward_test,
                    termination_test,
                    truncation_test,
                    info_test,
                ) = test_environment.step(actions_test)

            reward_train = reward_train.astype(numpy.float32, copy=False)
            reward_test = reward_test.astype(numpy.float32, copy=False)
            numpy.logical_or(
                termination_train,
                truncation_train,
                out=terminations[: self.train_environments],
            )
            numpy.logical_or(
                termination_test,
                truncation_test,
                out=terminations[self.train_environments :],
            )
            termination_train = terminations[: self.train_environments]

            score_rewards = self.__get_score_rewards(
                reward_train=reward_train,
                reward_test=reward_test,
                info_train=info_train,
                info_test=info_test,
                output=score_rewards_output,
            )
            completed_episodes = self.__get_episode_terminations(
                info_train=info_train,
                info_test=info_test,
                fallback_terminations=terminations,
                output=episode_terminations,
            )
            episode_returns += score_rewards
            self.__record_completed_episodes(
                episode_returns=episode_returns,
                terminations=completed_episodes,
            )
            self.__resample_terminated_bootstrap_heads(
                active_heads=active_heads,
                terminations=terminations,
            )

            bootstrap_masks = None
            if rollout_buffer.bootstrap_masks is not None:
                bootstrap_masks = (
                    full_bootstrap_masks
                    if full_bootstrap_masks is not None
                    else self.__sample_bootstrap_masks()
                )

            rollout_buffer.insert(
                step=step,
                observation=train_observation,
                action=actions_train_tensor,
                reward=self.__tensor(reward_train, torch.float32),
                termination=self.__tensor(termination_train, torch.float32),
                state_q_values=state_q_values,
                old_q_values=old_q_values,
                bootstrap_masks=bootstrap_masks,
            )

            self.__copy_array_to_tensor(
                observation[: self.train_environments],
                next_observation_train,
                torch.uint8,
            )
            self.__copy_array_to_tensor(
                observation[self.train_environments :],
                next_observation_test,
                torch.uint8,
            )
            frame_count += self.train_environments

        return observation, frame_count

    @torch.inference_mode()
    def __compute_targets(
        self,
        *,
        observation: torch.Tensor,
        rollout_buffer: RolloutBuffer,
    ):
        last_train_observation = observation[: self.train_environments]
        returns_rewards = rollout_buffer.rewards
        returns_terminations = rollout_buffer.terminations

        if self.__bootstrapped_enabled():
            with self.__autocast_float16():
                last_next_q_heads = self._network.get_q_heads(last_train_observation)
            last_next_q = last_next_q_heads.float().max(dim=-1).values
            returns_rewards = rollout_buffer.rewards.unsqueeze(-1).expand_as(
                rollout_buffer.state_q_values
            )
            returns_terminations = rollout_buffer.terminations.unsqueeze(-1).expand_as(
                rollout_buffer.state_q_values
            )
        else:
            with self.__autocast_float16():
                last_next_q_values = self.get_q_values(
                    float_observations=last_train_observation,
                    gradient=False,
                )
            last_next_q = last_next_q_values.float().max(dim=-1).values

        next_q = torch.empty_like(rollout_buffer.state_q_values)
        next_q[:-1] = rollout_buffer.state_q_values[1:]
        next_q[-1] = last_next_q

        return self.get_returns(
            batch_rewards=returns_rewards,
            batch_terminations=returns_terminations,
            next_q=next_q,
        )

    def __get_bootstrapped_loss(
        self,
        *,
        mini_batch_observations: torch.Tensor,
        mini_batch_actions: torch.Tensor,
        mini_batch_targets: torch.Tensor,
        mini_batch_bootstrap_masks: torch.Tensor,
        mini_batch_old_q_values: Optional[torch.Tensor] = None,
        mini_batch_target_probs: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if bool(getattr(self._network, "distributional", False)):
            return self.__get_distributional_bootstrapped_loss(
                mini_batch_observations=mini_batch_observations,
                mini_batch_actions=mini_batch_actions,
                mini_batch_targets=mini_batch_targets,
                mini_batch_bootstrap_masks=mini_batch_bootstrap_masks,
                mini_batch_old_q_values=mini_batch_old_q_values,
                mini_batch_target_probs=mini_batch_target_probs,
            )

        q_heads = self._network.get_q_heads(mini_batch_observations)
        action_indices = mini_batch_actions.reshape(-1, 1, 1).expand(
            -1,
            self.__get_bootstrap_heads(),
            1,
        )
        q_taken = q_heads.gather(2, action_indices).squeeze(2).float()
        loss = 0.5 * (q_taken - mini_batch_targets.float()).pow(2)
        masks = mini_batch_bootstrap_masks.to(dtype=loss.dtype)
        return (loss * masks).sum() / masks.sum().clamp_min(1.0)

    def __get_distributional_bootstrapped_cross_entropy(
        self,
        *,
        logits: torch.Tensor,
        target_probs: torch.Tensor,
    ) -> torch.Tensor:
        bins = logits.shape[-1]
        flattened_logits = logits.reshape(-1, bins)
        flattened_target_probs = target_probs.reshape(-1, bins)
        if logits.device.type == "cuda":
            with torch.autocast("cuda", enabled=False):
                loss = torch.nn.functional.cross_entropy(
                    flattened_logits,
                    flattened_target_probs,
                    reduction="none",
                )
        else:
            loss = torch.nn.functional.cross_entropy(
                flattened_logits,
                flattened_target_probs,
                reduction="none",
            )
        return loss.reshape(logits.shape[:-1])

    def __get_distributional_bootstrapped_loss(
        self,
        *,
        mini_batch_observations: torch.Tensor,
        mini_batch_actions: torch.Tensor,
        mini_batch_targets: torch.Tensor,
        mini_batch_bootstrap_masks: torch.Tensor,
        mini_batch_old_q_values: Optional[torch.Tensor],
        mini_batch_target_probs: Optional[torch.Tensor],
    ) -> torch.Tensor:
        q_logits_heads = self._network.get_q_logits_heads(mini_batch_observations)
        action_indices = mini_batch_actions.reshape(-1, 1, 1, 1).expand(
            -1,
            self.__get_bootstrap_heads(),
            1,
            q_logits_heads.shape[-1],
        )
        q_logits_taken = q_logits_heads.gather(2, action_indices).squeeze(2)

        if mini_batch_target_probs is None:
            bins = q_logits_taken.shape[-1]
            loss = self._network.hl_gauss_loss(
                q_logits_taken.reshape(-1, bins),
                mini_batch_targets.reshape(-1),
                reduction="none",
            ).reshape(q_logits_taken.shape[:-1])
        else:
            loss = self.__get_distributional_bootstrapped_cross_entropy(
                logits=q_logits_taken,
                target_probs=mini_batch_target_probs,
            )

        value_clip = float(getattr(self, "distributional_value_clip", 0.0))
        if mini_batch_old_q_values is not None and value_clip > 0.0:
            bins = q_logits_taken.shape[-1]
            scalar_q_taken = self._network.hl_gauss_loss(
                q_logits_taken.reshape(-1, bins),
            ).reshape(q_logits_taken.shape[:-1])
            scalar_q_clipped = mini_batch_old_q_values + (
                scalar_q_taken - mini_batch_old_q_values
            ).clamp(-value_clip, value_clip)
            q_clipped_logprobs = self._network.hl_gauss_loss.transform_to_logprobs(
                scalar_q_clipped.reshape(-1)
            ).reshape(*scalar_q_clipped.shape, -1)
            if mini_batch_target_probs is None:
                clipped_loss = self._network.hl_gauss_loss(
                    q_clipped_logprobs.reshape(-1, bins),
                    mini_batch_targets.reshape(-1),
                    reduction="none",
                ).reshape(q_logits_taken.shape[:-1])
            else:
                clipped_loss = self.__get_distributional_bootstrapped_cross_entropy(
                    logits=q_clipped_logprobs,
                    target_probs=mini_batch_target_probs,
                )
            loss = torch.max(loss, clipped_loss)

        masks = mini_batch_bootstrap_masks.to(dtype=loss.dtype)
        return (loss * masks).sum() / masks.sum().clamp_min(1.0)

    def __slice_optional(
        self,
        tensor: Optional[torch.Tensor],
        indices: torch.Tensor,
    ) -> Optional[torch.Tensor]:
        return None if tensor is None else tensor[indices]

    def __update_network(
        self,
        *,
        flattened_observations: torch.Tensor,
        flattened_actions: torch.Tensor,
        flattened_old_q_values: Optional[torch.Tensor],
        flattened_bootstrap_masks: Optional[torch.Tensor],
        flattened_targets: torch.Tensor,
        flattened_target_probs: Optional[torch.Tensor],
        scaler,
    ):
        self._network.train()
        log_batches = bool(getattr(self, "loss_log_batches", False))
        update_losses = [] if log_batches else None
        update_loss_sum = None
        update_loss_count = 0
        clip_grad_foreach = self.device.type in {"cpu", "cuda"}
        for _ in range(self.epochs):
            indices = torch.randperm(self.batch_size, device=self.device)

            for mini_batch_idx in indices.split(self.mini_batch_size):
                mini_batch_observations = flattened_observations[mini_batch_idx]
                mini_batch_actions = flattened_actions[mini_batch_idx]
                mini_batch_targets = flattened_targets[mini_batch_idx]
                mini_batch_old_q_values = self.__slice_optional(
                    flattened_old_q_values,
                    mini_batch_idx,
                )
                mini_batch_target_probs = self.__slice_optional(
                    flattened_target_probs,
                    mini_batch_idx,
                )
                mini_batch_bootstrap_masks = self.__slice_optional(
                    flattened_bootstrap_masks,
                    mini_batch_idx,
                )

                self._optimizer.zero_grad(set_to_none=True)
                with self.__autocast_float16():
                    if mini_batch_bootstrap_masks is None:
                        loss = self.get_loss(
                            mini_batch_observations=mini_batch_observations,
                            mini_batch_targets=mini_batch_targets,
                            mini_batch_actions=mini_batch_actions,
                            mini_batch_old_q_values=mini_batch_old_q_values,
                            mini_batch_target_probs=mini_batch_target_probs,
                        )
                    else:
                        loss = self.__get_bootstrapped_loss(
                            mini_batch_observations=mini_batch_observations,
                            mini_batch_targets=mini_batch_targets,
                            mini_batch_actions=mini_batch_actions,
                            mini_batch_bootstrap_masks=mini_batch_bootstrap_masks,
                            mini_batch_old_q_values=mini_batch_old_q_values,
                            mini_batch_target_probs=mini_batch_target_probs,
                        )
                scaler.scale(loss).backward()
                scaler.unscale_(self._optimizer)
                torch.nn.utils.clip_grad_norm_(
                    self._network.parameters(),
                    self.gradient_norm,
                    foreach=clip_grad_foreach,
                )
                scaler.step(self._optimizer)
                scaler.update()
                detached_loss = loss.detach()
                if log_batches:
                    update_losses.append(detached_loss)
                else:
                    update_loss_sum = (
                        detached_loss
                        if update_loss_sum is None
                        else update_loss_sum + detached_loss
                    )
                    update_loss_count += 1

        if log_batches and update_losses:
            self.results.loss.extend(torch.stack(update_losses).float().cpu().tolist())
        elif update_loss_sum is not None and update_loss_count > 0:
            mean_loss = update_loss_sum / update_loss_count
            self.results.loss.append(float(mean_loss.float().cpu().item()))

    def __log_progress(self, *, update: int, frame_count: int):
        if update % self.verbose_interval != 0:
            return

        verbose_window = self.verbose_window
        test_score = (
            0.0
            if len(self.results.rewards.test) < verbose_window
            else numpy.mean(self.results.rewards.test[-verbose_window:])
        )

        self.flush_verbose(
            f"Update {update} | Frames: {frame_count * self.frame_skip:,}"
        )
        self.flush_verbose(f"Test Score: {test_score:.4f}")

    def _train(self, *, environment: str, seed: int):
        frame_count = 0
        (
            train_environment,
            test_environment,
            observation_shape,
            scaler,
            observation,
            episode_returns,
        ) = self.__initialize_training(environment=environment, seed=seed)

        rollout_buffer = self.__allocate_buffers(
            observation_shape=observation_shape,
        )
        active_heads = (
            self.__sample_bootstrap_heads(self.total_environments)
            if self.__bootstrapped_enabled()
            else None
        )

        training_start_time = time.time()

        for update in range(1, self.total_updates + 1):
            self._network.eval()

            observation, frame_count = self.__collect_trajectories(
                frame_count=frame_count,
                observation=observation,
                train_environment=train_environment,
                test_environment=test_environment,
                episode_returns=episode_returns,
                rollout_buffer=rollout_buffer,
                active_heads=active_heads,
            )

            targets = self.__compute_targets(
                observation=observation,
                rollout_buffer=rollout_buffer,
            )

            (
                flattened_observations,
                flattened_actions,
                flattened_old_q_values,
                flattened_targets,
                flattened_bootstrap_masks,
            ) = rollout_buffer.flatten(targets)
            flattened_target_probs = None
            if bool(getattr(self._network, "distributional", False)):
                target_shape = flattened_targets.shape
                flattened_target_probs = self._network.hl_gauss_loss.transform_to_probs(
                    flattened_targets.reshape(-1)
                ).reshape(*target_shape, -1)

            self.__update_network(
                scaler=scaler,
                flattened_observations=flattened_observations,
                flattened_actions=flattened_actions,
                flattened_old_q_values=flattened_old_q_values,
                flattened_bootstrap_masks=flattened_bootstrap_masks,
                flattened_targets=flattened_targets,
                flattened_target_probs=flattened_target_probs,
            )

            self.__log_progress(update=update, frame_count=frame_count)

        train_environment.close()
        if test_environment is not None:
            test_environment.close()
        self.results.duration = time.time() - training_start_time



from aftab.maps import optimizer_map
from .AftabBaseMixin import AftabBaseMixin


class AftabOptimizerMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    def _initialize_optimizer(self):
        if self.optimizer not in optimizer_map:
            raise ValueError(f"Optimizer `{self.optimizer}` was not founded.")

        self._optimizer = optimizer_map[self.optimizer](
            self._network.parameters(),
            lr=self.lr,
            eps=self.optimizer_epsilon,
            betas=(self.optimizer_first_beta, self.optimizer_second_beta),
            weight_decay=self.optimizer_weight_decay,
        )



import jax
import flax.linen as nn
from typing import Callable
from .LayerNorm2d import LayerNorm2d


class EncoderBlock(nn.Module):
    out_channels: int
    kernel: int
    stride: int
    padding: int | str = "SAME"
    activation: Callable = nn.relu

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        x = nn.Conv(
            features=self.out_channels,
            kernel_size=(self.kernel, self.kernel),
            strides=(self.stride, self.stride),
            padding=self.padding,
        )(x)
        x = LayerNorm2d()(x)
        x = self.activation(x)
        return x



import jax
import flax.linen as nn
from aftab.jax.modules import Stream


class QNetowrk(nn.Module):
    encoder: nn.Module
    action_dimension: int

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        features = self.encoder(x)
        return Stream(512, self.action_dimension)(features)



import torch
from typing import Optional
from pathlib import Path


def _make_sure_directory_exists(directory: str):
    directory = directory.replace(".", "/")
    Path(directory).mkdir(exist_ok=True, parents=True)
    return directory


class LinearEpsilon:
    def __init__(self, span: float = 0.1, minimum: float = 0.001, maximum: float = 1.0):
        self.maximum = maximum
        self.minimum = minimum
        self.span = span

    def get(self, frames, total_frames):
        minimum = self.minimum
        maximum = self.maximum
        decay_duration = total_frames * self.span

        if decay_duration == 0:
            return maximum

        delta = maximum - minimum
        return max(minimum, maximum - (frames / decay_duration) * delta)


class RolloutBuffer:
    def __init__(
        self,
        *,
        observation_shape: tuple,
        steps_per_update: int,
        train_environments: int,
        device: torch.device,
        bootstrapped: bool = False,
        bootstrap_heads: int = 1,
        store_old_q_values: bool = False,
    ):
        rollout_shape = (steps_per_update, train_environments)
        state_q_shape = rollout_shape
        if bootstrapped:
            state_q_shape = rollout_shape + (bootstrap_heads,)

        self.observations = torch.empty(
            rollout_shape + observation_shape,
            dtype=torch.uint8,
            device=device,
        )
        self.actions = torch.empty(
            rollout_shape,
            dtype=torch.int64,
            device=device,
        )
        self.rewards = torch.empty(
            rollout_shape,
            dtype=torch.float32,
            device=device,
        )
        self.terminations = torch.empty(
            rollout_shape,
            dtype=torch.float32,
            device=device,
        )
        self.state_q_values = torch.empty(
            state_q_shape,
            dtype=torch.float32,
            device=device,
        )
        self.old_q_values = None
        if store_old_q_values:
            old_q_shape = state_q_shape if bootstrapped else rollout_shape
            self.old_q_values = torch.empty(
                old_q_shape,
                dtype=torch.float32,
                device=device,
            )
        self.bootstrap_masks = None
        if bootstrapped:
            self.bootstrap_masks = torch.empty(
                state_q_shape,
                dtype=torch.float32,
                device=device,
            )

    def insert(
        self,
        *,
        step: int,
        observation: torch.Tensor,
        action: torch.Tensor,
        reward: torch.Tensor,
        termination: torch.Tensor,
        state_q_values: torch.Tensor,
        old_q_values: Optional[torch.Tensor] = None,
        bootstrap_masks: Optional[torch.Tensor] = None,
    ) -> None:
        self.observations[step] = observation
        self.actions[step] = action
        self.rewards[step] = reward
        self.terminations[step] = termination
        self.state_q_values[step] = state_q_values

        if self.old_q_values is not None:
            if old_q_values is None:
                raise ValueError("Expected `old_q_values` for this rollout buffer.")
            self.old_q_values[step] = old_q_values

        if self.bootstrap_masks is not None:
            if bootstrap_masks is None:
                raise ValueError("Expected `bootstrap_masks` for this rollout buffer.")
            self.bootstrap_masks[step] = bootstrap_masks

    def flatten(self, targets: torch.Tensor):
        flattened_observations = self.observations.flatten(0, 1)
        flattened_actions = self.actions.reshape(-1)
        flattened_old_q_values = None
        if self.old_q_values is not None:
            flattened_old_q_values = self.old_q_values.flatten(0, 1)
        flattened_targets = targets.flatten(0, 1)
        flattened_bootstrap_masks = None
        if self.bootstrap_masks is not None:
            flattened_bootstrap_masks = self.bootstrap_masks.flatten(0, 1)
        return (
            flattened_observations,
            flattened_actions,
            flattened_old_q_values,
            flattened_targets,
            flattened_bootstrap_masks,
        )



import torch
import inspect
from aftab.constants import ModuleType


class Stream(torch.nn.Module):
    def __init__(
        self,
        *,
        input_dimension: int,
        hidden_dimension: int,
        output_dimension: int,
        activation: ModuleType = torch.nn.ReLU,
        normalization: bool = True,
    ):
        super().__init__()
        self.normalization = normalization

        self.first_linear = torch.nn.Linear(input_dimension, hidden_dimension)
        self.second_linear = torch.nn.Linear(hidden_dimension, output_dimension)

        self.activation = self.__initiate_activation(
            activation_class=activation, channels=hidden_dimension
        )

        if normalization:
            self.normalization_layer = torch.nn.LayerNorm(hidden_dimension)

    def __initiate_activation(
        self, activation_class: ModuleType, channels: int
    ) -> torch.nn.Module:
        activation_args = {}
        args = inspect.signature(activation_class).parameters

        if "inplace" in args:
            activation_args.update({"inplace": True})

        if "channels" in args:
            activation_args.update({"channels": channels})
        elif "in_channels" in args:
            activation_args.update({"in_channels": channels})

        return activation_class(**activation_args)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.first_linear(x)
        if self.normalization:
            x = self.normalization_layer(x)
        x = self.activation(x)
        x = self.second_linear(x)
        return x



import jax
import flax.linen as nn


class Linear(nn.Module):
    input_dimension: int
    output_dimension: int

    def __call__(self, x: jax.Array) -> jax.Array:
        return nn.Dense(self.output_dimension)(x)



import jax
import jax.numpy as jnp
import flax.linen as nn
from flax.training.train_state import TrainState
import optax
import envpool
from typing import Any, NamedTuple


# ---------------------------------------------------------
# 1. Dataclasses & Replay Buffer Transition
# ---------------------------------------------------------
class Transition(NamedTuple):
    obs: jnp.ndarray
    action: jnp.ndarray
    reward: jnp.ndarray
    done: jnp.ndarray
    next_obs: jnp.ndarray
    q_val: jnp.ndarray


class CustomTrainState(TrainState):
    batch_stats: Any


# ---------------------------------------------------------
# 2. Neural Networks (Flax -> equivalent to PyTorch nn.Module)
# ---------------------------------------------------------
class CNN(nn.Module):
    norm_type: str = "layer_norm"

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool):
        def normalize(inputs):
            if self.norm_type == "layer_norm":
                return nn.LayerNorm()(inputs)
            elif self.norm_type == "batch_norm":
                return nn.BatchNorm(use_running_average=not train)(inputs)
            return inputs

        # PyTorch style forward pass
        x = nn.Conv(
            32,
            kernel_size=(8, 8),
            strides=(4, 4),
            padding="VALID",
            kernel_init=nn.initializers.he_normal(),
        )(x)
        x = normalize(x)
        x = nn.relu(x)

        x = nn.Conv(
            64,
            kernel_size=(4, 4),
            strides=(2, 2),
            padding="VALID",
            kernel_init=nn.initializers.he_normal(),
        )(x)
        x = normalize(x)
        x = nn.relu(x)

        x = nn.Conv(
            64,
            kernel_size=(3, 3),
            strides=(1, 1),
            padding="VALID",
            kernel_init=nn.initializers.he_normal(),
        )(x)
        x = normalize(x)
        x = nn.relu(x)

        x = x.reshape((x.shape[0], -1))  # Flatten
        x = nn.Dense(512, kernel_init=nn.initializers.he_normal())(x)
        x = normalize(x)
        x = nn.relu(x)
        return x


class QNetwork(nn.Module):
    action_dim: int
    norm_type: str = "layer_norm"
    norm_input: bool = False

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool):
        # NHWC format for JAX standard
        x = jnp.transpose(x, (0, 2, 3, 1))
        if self.norm_input:
            x = nn.BatchNorm(use_running_average=not train)(x)
        else:
            x_dummy = nn.BatchNorm(use_running_average=not train)(x)
            x = x / 255.0

        x = CNN(norm_type=self.norm_type)(x, train)
        x = nn.Dense(self.action_dim)(x)
        return x


# ---------------------------------------------------------
# 3. JIT-Compiled Training Components
# ---------------------------------------------------------


@jax.jit
def select_action(
    train_state: CustomTrainState,
    obs: jnp.ndarray,
    eps: jnp.ndarray,
    rng: jax.random.PRNGKey,
):
    """Epsilon-greedy action selection."""
    rng_a, rng_e = jax.random.split(rng)

    # Forward pass (eval mode)
    q_vals = train_state.apply_fn(
        {"params": train_state.params, "batch_stats": train_state.batch_stats},
        obs,
        train=False,
    )

    greedy_actions = jnp.argmax(q_vals, axis=-1)
    random_actions = jax.random.randint(
        rng_a, greedy_actions.shape, minval=0, maxval=q_vals.shape[-1]
    )

    # Where random condition is met, pick random_action, else greedy
    is_random = jax.random.uniform(rng_e, greedy_actions.shape) < eps
    actions = jnp.where(is_random, random_actions, greedy_actions)

    return actions, q_vals


@jax.jit
def compute_targets(last_q, q_vals, rewards, dones, gamma, lambda_):
    """Computes lambda returns (GAE-style) running backwards."""

    def _get_target(carry, transition):
        lambda_returns, next_q = carry
        reward, q, done = transition

        target_bootstrap = reward + gamma * (1 - done) * next_q
        delta = lambda_returns - next_q
        lambda_returns = target_bootstrap + gamma * lambda_ * delta
        lambda_returns = (1 - done) * lambda_returns + done * reward
        next_q = jnp.max(q, axis=-1)

        return (lambda_returns, next_q), lambda_returns

    lambda_returns = rewards[-1] + gamma * (1 - dones[-1]) * last_q
    last_q = jnp.max(q_vals[-1], axis=-1)

    # Scan backwards over the sequence (excluding the very last step initialized above)
    _, targets = jax.lax.scan(
        _get_target,
        (lambda_returns, last_q),
        (rewards[:-1], q_vals[:-1], dones[:-1]),
        reverse=True,
    )

    # Append the last step back to the targets
    targets = jnp.concatenate([targets, lambda_returns[None, ...]])
    return targets


@jax.jit
def train_step(
    train_state: CustomTrainState,
    obs: jnp.ndarray,
    actions: jnp.ndarray,
    targets: jnp.ndarray,
):
    """A single PyTorch-style minibatch gradient update."""

    def loss_fn(params):
        q_vals, updates = train_state.apply_fn(
            {"params": params, "batch_stats": train_state.batch_stats},
            obs,
            train=True,
            mutable=["batch_stats"],
        )

        # Gather Q-values for the actions we actually took: Q(s, a)
        chosen_action_qvals = jnp.take_along_axis(
            q_vals, jnp.expand_dims(actions, axis=-1), axis=-1
        ).squeeze(axis=-1)

        loss = 0.5 * jnp.square(chosen_action_qvals - targets).mean()
        return loss, updates

    # Compute gradients and updates
    (loss, updates), grads = jax.value_and_grad(loss_fn, has_aux=True)(
        train_state.params
    )

    # Apply gradients (PyTorch optimizer.step())
    train_state = train_state.apply_gradients(grads=grads)
    train_state = train_state.replace(batch_stats=updates["batch_stats"])

    return train_state, loss


# ---------------------------------------------------------
# 4. Main Training Loop (PyTorch Style)
# ---------------------------------------------------------
def main():
    # --- Configuration ---
    config = {
        "ENV_NAME": "Pong-v5",
        "SEED": 42,
        "NUM_ENVS": 8,
        "TEST_ENVS": 2,
        "TEST_DURING_TRAINING": True,
        "NUM_STEPS": 128,  # Steps per rollout
        "NUM_MINIBATCHES": 4,  # Minibatches per epoch
        "NUM_EPOCHS": 3,  # Epochs per PPO/PQN update
        "TOTAL_TIMESTEPS": 10_000_000,
        "LR": 3e-4,
        "MAX_GRAD_NORM": 0.5,
        "GAMMA": 0.99,
        "LAMBDA": 0.95,
        "EPS_START": 1.0,
        "EPS_FINISH": 0.05,
        "EPS_DECAY": 0.5,
        "NORM_TYPE": "layer_norm",
    }

    # Derived sizes
    total_envs = (
        config["NUM_ENVS"] + config["TEST_ENVS"]
        if config["TEST_DURING_TRAINING"]
        else config["NUM_ENVS"]
    )
    num_updates = config["TOTAL_TIMESTEPS"] // config["NUM_STEPS"] // config["NUM_ENVS"]
    batch_size = config["NUM_STEPS"] * config["NUM_ENVS"]
    minibatch_size = batch_size // config["NUM_MINIBATCHES"]

    # --- Setup Env ---
    env = envpool.make(
        config["ENV_NAME"], env_type="gym", num_envs=total_envs, seed=config["SEED"]
    )
    obs, env_state = env.reset()

    # --- Setup Agent & Optimizer ---
    rng = jax.random.PRNGKey(config["SEED"])
    rng, init_rng = jax.random.split(rng)

    network = QNetwork(action_dim=env.action_space.n, norm_type=config["NORM_TYPE"])
    init_obs = jnp.zeros((1, *env.observation_space.shape))
    variables = network.init(init_rng, init_obs, train=False)

    optimizer = optax.chain(
        optax.clip_by_global_norm(config["MAX_GRAD_NORM"]),
        optax.radam(learning_rate=config["LR"]),
    )

    train_state = CustomTrainState.create(
        apply_fn=network.apply,
        params=variables["params"],
        batch_stats=variables["batch_stats"],
        tx=optimizer,
    )

    # Epsilon scheduler setup
    eps_decay_steps = int(config["EPS_DECAY"] * num_updates)
    eps_scheduler = optax.linear_schedule(
        config["EPS_START"], config["EPS_FINISH"], eps_decay_steps
    )

    # --- PyTorch-style Training Loop ---
    for update in range(num_updates):
        transitions = []

        # ---------------------------------------------------
        # Phase 1: Environment Rollouts
        # ---------------------------------------------------
        for step in range(config["NUM_STEPS"]):
            rng, action_rng = jax.random.split(rng)

            # Assign epsilon: Exploration for train envs, 0.0 for test envs
            current_eps = eps_scheduler(update)
            eps_array = jnp.full(config["NUM_ENVS"], current_eps)
            if config["TEST_DURING_TRAINING"]:
                eps_array = jnp.concatenate((eps_array, jnp.zeros(config["TEST_ENVS"])))

            # Step agent
            actions, q_vals = select_action(train_state, obs, eps_array, action_rng)
            next_obs, next_env_state, reward, done, info = env.step(env_state, actions)

            # Store transition
            transitions.append(Transition(obs, actions, reward, done, next_obs, q_vals))

            obs = next_obs
            env_state = next_env_state

        # Stack list of transitions into unified JAX arrays (PyTorch `torch.stack` equivalent)
        batch = jax.tree_util.tree_map(lambda *x: jnp.stack(x), *transitions)

        # ---------------------------------------------------
        # Phase 2: Compute Targets
        # ---------------------------------------------------
        # Discard test envs for gradient calculation targets
        if config["TEST_DURING_TRAINING"]:
            train_batch = jax.tree_util.tree_map(
                lambda x: x[:, : -config["TEST_ENVS"]], batch
            )
        else:
            train_batch = batch

        # Get Q-value of the final state for bootstrapping
        last_q_vals = train_state.apply_fn(
            {"params": train_state.params, "batch_stats": train_state.batch_stats},
            train_batch.next_obs[-1],
            train=False,
        )
        last_q = jnp.max(last_q_vals, axis=-1)

        targets = compute_targets(
            last_q,
            train_batch.q_val,
            train_batch.reward,
            train_batch.done,
            config["GAMMA"],
            config["LAMBDA"],
        )

        # ---------------------------------------------------
        # Phase 3: Optimize / Epoch Loop
        # ---------------------------------------------------
        # Flatten time and env dimensions: (Num_Steps, Num_Envs, ...) -> (Batch_Size, ...)
        flat_obs = train_batch.obs.reshape(-1, *train_batch.obs.shape[2:])
        flat_actions = train_batch.action.reshape(-1)
        flat_targets = targets.reshape(-1)

        dataset_size = flat_obs.shape[0]

        for epoch in range(config["NUM_EPOCHS"]):
            rng, shuffle_rng = jax.random.split(rng)

            # Shuffle indices (Standard PyTorch dataloader behavior)
            indices = jax.random.permutation(shuffle_rng, dataset_size)

            for start_idx in range(0, dataset_size, minibatch_size):
                batch_idx = indices[start_idx : start_idx + minibatch_size]

                # Sample minibatches
                mb_obs = flat_obs[batch_idx]
                mb_actions = flat_actions[batch_idx]
                mb_targets = flat_targets[batch_idx]

                # Gradient step
                train_state, loss = train_step(
                    train_state, mb_obs, mb_actions, mb_targets
                )

        # Basic Logging
        if update % 10 == 0:
            print(
                f"Update: {update}/{num_updates} | Loss: {loss:.4f} | Eps: {current_eps:.3f}"
            )


if __name__ == "__main__":
    main()



from baloot import funnel
from aftab.common import _make_sure_directory_exists
from .AftabBaseMixin import AftabBaseMixin


class AftabTrainingResultsMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    def __make_log_filename(self) -> str:
        filename = f"experiment-{self.experiment_name}__"
        filename += f"network-{self.network}__"
        filename += f"seed-{self.buffer.seed}__"
        filename += f"environment-{self.buffer.environment}__"
        filename += f"encoder-{self.encoder.__name__}__"
        # removes trailing __
        filename = filename.strip("__")
        return f"{filename}.pkl"

    def __build_log_payload(self) -> dict:
        duration = self.results.duration or 0
        data = {
            "environment": self.buffer.environment,
            "seed": self.buffer.seed,
            "encoder": self.encoder.__name__,
            "optimizer": self.optimizer,
            "lambda": self.return_lambda,
            "network": self.network,
            "frames": self.frames,
            "frame_skip": self.frame_skip,
            "training_reward": self.results.rewards.train,
            "test_reward": self.results.rewards.test,
            "loss": self.results.loss,
            "duration_seconds": duration,
            "duration_hours": duration / 3600,
        }

        if self.network in [
            "distributional",
            "distributional-duelling",
            "distributional-bootstrapped-duelling",
        ]:
            data.update(
                {
                    "distributional_bins": self.distributional_bins,
                    "distributional_min_value": self.distributional_min_value,
                    "distributional_max_value": self.distributional_max_value,
                    "distributional_sigma": self.distributional_sigma,
                    "distributional_sigma_ratio": self.distributional_sigma_ratio,
                    "distributional_value_clip": self.distributional_value_clip,
                }
            )

        if self.network in [
            "bootstrapped",
            "bootstrapped-duelling",
            "distributional-bootstrapped-duelling",
        ]:
            data.update(
                {
                    "bootstrap_heads": self.bootstrap_heads,
                    "bootstrap_probability": self.bootstrap_probability,
                }
            )

        return data

    def _log(self, *, directory: str) -> None:
        directory_path = _make_sure_directory_exists(directory).strip("/").strip()
        filename = self.__make_log_filename()
        payload = self.__build_log_payload()
        funnel(f"{directory_path}/{filename}", payload)



import torch
from aftab.modules import EncoderBlock
from aftab.constants import ModuleType


class AlphaEncoder(torch.nn.Module):
    def __init__(self, *, activation: ModuleType = torch.nn.ReLU):
        super().__init__()
        self.stream = torch.nn.Sequential(
            EncoderBlock(
                4, 32, kernel_size=4, stride=2, padding=1, activation=activation
            ),
            EncoderBlock(
                32, 64, kernel_size=4, stride=2, padding=1, activation=activation
            ),
            EncoderBlock(
                64, 64, kernel_size=3, stride=2, padding=1, activation=activation
            ),
            EncoderBlock(
                64, 64, kernel_size=5, stride=1, padding=0, activation=activation
            ),
            torch.nn.Flatten(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.stream(x)



import jax
import flax.linen as nn
from aftab.jax.modules import Conv2d


class NatureDQNEncoder(nn.Module):
    @nn.compact
    def __call__(self, x: jax.Array):
        x = Conv2d(4, 32, 8, 4)(x)
        x = nn.relu(x)
        x = Conv2d(32, 64, 4, 2)(x)
        x = nn.relu(x)
        x = Conv2d(64, 64, 3, 1)(x)
        x = nn.relu(x)
        return x



import numpy
import torch


def epsilon_greedy_vectorized(q_values, eps):
    if isinstance(eps, numpy.ndarray):
        if eps.dtype != numpy.float32:
            eps = eps.astype(numpy.float32, copy=False)
        if not eps.flags.c_contiguous:
            eps = numpy.ascontiguousarray(eps)
        eps = torch.from_numpy(eps).to(
            device=q_values.device,
            non_blocking=q_values.device.type == "cuda",
        )
    elif isinstance(eps, list):
        eps = torch.from_numpy(numpy.asarray(eps, dtype=numpy.float32)).to(
            device=q_values.device,
            non_blocking=q_values.device.type == "cuda",
        )
    elif isinstance(eps, (float, int, numpy.number)):
        eps = float(eps)
    elif isinstance(eps, torch.Tensor):
        eps = eps.to(q_values.device)

    num_envs, action_dim = q_values.shape
    if isinstance(eps, float):
        if eps <= 0.0:
            return torch.argmax(q_values, dim=-1)
        if eps >= 1.0:
            return torch.randint(0, action_dim, (num_envs,), device=q_values.device)

    if (
        isinstance(eps, torch.Tensor)
        and eps.ndim == 1
        and eps.shape[0] != q_values.shape[0]
    ):
        eps = eps[0]

    greedy_actions = torch.argmax(q_values, dim=-1)
    random_actions = torch.randint(0, action_dim, (num_envs,), device=q_values.device)

    mask = torch.rand(num_envs, device=q_values.device) < eps
    final_actions = torch.where(mask, random_actions, greedy_actions)
    return final_actions



import argparse
from aftab import Aftab

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--seed",
        type=int,
        required=True,
        help="Aftab running seed"
    )
    parser.add_argument(
        "--environment",
        type=str,
        required=True,
        help="Environment name"
    )
    args = parser.parse_args()
    agent = Aftab(verbose=True)
    agent.train(environment=args.environment, seed=args.seed)
    agent.log(directory="results")

if __name__ == "__main__":
    main()


import torch
from aftab.mixins import AftabBaseMixin
from aftab.common import _make_sure_directory_exists


class AftabStoreModelMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    def _save(self, directory: str):
        directory_path = _make_sure_directory_exists(directory).strip("/").strip()
        filename = f"{directory_path}/{self.__make_network_filename()}.model"
        torch.save(self._network, filename)



from __future__ import annotations

import torch
import math
import os
import time
from typing import Literal
from types import SimpleNamespace
from baloot import acceleration_device, seed_everything
from aftab.constants import (
    ModuleType,
    EncoderStringType,
    OptimizerStringType,
    NetworkStringType,
)
from aftab.maps import encoders_map, acceptable_frames_map
from aftab.functions import flush
from aftab.mixins import *


class Aftab(
    AftabOptimizerMixin,
    AftabTrainingResultsMixin,
    AftabEnvironmentMixin,
    AftabActionsMixin,
    AftabNetworkMixin,
    AftabQValueMixin,
    AftabLossMixin,
    AftabReturnsMixin,
    AftabTrainMixin,
    AftabStoreModelMixin,
):
    def __init__(
        self,
        *,
        experiment_name: str,
        encoder: ModuleType | EncoderStringType = "gammahadamaxv1",
        network: NetworkStringType = "distributional-bootstrapped-duelling",
        frames: int | Literal["pilot", "full", "ablation"] = "full",
        frame_skip: int = 4,
        mini_batches: int = 32,
        epochs: int = 2,
        gamma: float = 0.99,
        return_lambda: float = 0.65,
        lr: float = 25e-5,
        train_environments: int = 128,
        test_environments: int = 8,
        steps_per_update: int = 32,
        min_cpu_count: int = 4,
        noop: int = 30,
        frame_stack: int = 4,
        gradient_norm: float = 10.0,
        verbose: bool = False,
        verbose_interval: int = 10,
        verbose_window: int = 100,
        embedding_dimension: int = 512,
        optimizer: OptimizerStringType = "radam",
        optimizer_epsilon: float = 1e-5,
        optimizer_weight_decay: float = 0.0,
        optimizer_first_beta: float = 0.9,
        optimizer_second_beta: float = 0.999,
        autocast_float16: bool = True,
        channels_last: bool = True,
        train_episodic_life: bool = True,
        test_episodic_life: bool = False,
        train_reward_clip: bool = True,
        test_reward_clip: bool = True,
        loss_log_batches: bool = False,
        distributional_bins: int = 51,
        distributional_min_value: float = -10.0,
        distributional_max_value: float = 10.0,
        distributional_sigma: float | None = None,
        distributional_sigma_ratio: float = 0.75,
        distributional_value_clip: float = 0.0,
        bootstrap_heads: int = 10,
        bootstrap_probability: float = 1.0,
        seed_offset: int = 1000,
    ):
        params = locals()
        params.pop("self")
        self.__initialize(params)
        super().__init__()

    def __initialize(self, params):
        self.__initialize_hyperparameters(**params)
        self.__initialize_frames()
        self.__initialize_derived_attributes()
        self.__initialize_constants()
        self.__initialize__encoder()
        self.__initialize_buffer()

    def __initialize_buffer(self) -> None:
        self.buffer = SimpleNamespace()

    def __initialize_hyperparameters(self, **hyperparameters) -> None:
        for key, value in hyperparameters.items():
            setattr(self, key, value)

    def __initialize_frames(self):
        if isinstance(self.frames, int):
            return

        if self.frames not in acceptable_frames_map:
            raise ValueError(f"Frames was provided a wrong value `{self.frames}`")

        self.frames = acceptable_frames_map[self.frames]

    def __initialize_derived_attributes(self):
        self.total_environments = int(self.train_environments + self.test_environments)
        self.batch_size = int(self.train_environments * self.steps_per_update)
        self.mini_batch_size = int(self.batch_size / self.mini_batches)
        self.effective_frames = int(self.frames / self.frame_skip)
        self.total_updates = math.ceil(self.effective_frames / self.batch_size)

    def __initialize__encoder(self):
        if isinstance(self.encoder, str):
            if self.encoder not in encoders_map:
                raise ValueError(f"Encoder was provided a wrong value `{self.encoder}`")
            self.encoder = encoders_map[self.encoder]

        self.flush_verbose(f"Encoder: {self.encoder.__name__}")

    def __initialize_constants(self):
        self.device = acceleration_device()
        self.cpu_count = os.cpu_count() or 1
        self.flush_verbose(f"Acceleration device: {self.device}")
        self.flush_verbose(f"CPU Count: {self.cpu_count}")

    def __set_precision(self):
        torch.set_float32_matmul_precision("high")

    def __set_seed(self, seed: int):
        seed_everything(seed)

    def __set_buffer(self, key, value):
        setattr(self.buffer, key, value)

    def __clear_results(self):
        self.results = SimpleNamespace()
        self.results.rewards = SimpleNamespace()
        self.results.rewards.train = []
        self.results.rewards.test = []
        self.results.loss = []
        self.results.duration = 0.0

    def __make_network_filename(self) -> str:
        timestamp = time.time()
        filename = (
            f"{timestamp}__{self.experiment_name}__{self.encoder}__{self.network}.model"
        )
        return filename

    def flush_verbose(self, message: str):
        if not self.verbose:
            return

        flush(message=message)

    def save(self, directory: str = "models"):
        self._save(directory=directory)

    def log(self, directory: str = "results") -> None:
        self._log(directory=directory)

    def train(self, *, environment: str, seed: int):
        self.__set_precision()
        self.__set_seed(seed)
        self.__clear_results()
        self.__set_buffer("seed", seed)
        self.__set_buffer("environment", environment)
        self.flush_verbose(f"Environment: {environment}")
        self.flush_verbose(f"Seed: {seed}")
        self._train(environment=environment, seed=seed)



import jax
import flax.linen as nn


class Conv2d(nn.Module):
    in_channels: int
    out_channels: int
    kernel_size: int
    stride: int
    padding: int | str = "SAME"

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        return nn.Conv(self.out_channels, self.kernel_size, self.stride, self.padding)(
            x
        )



import jax
import flax.linen as nn


class Stream(nn.Module):
    hidden_dimension: int
    output_dimension: int

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        x = nn.Dense(self.hidden_dimension)(x)
        x = nn.relu(x)
        x = nn.Dense(self.output_dimension)(x)
        return x



import torch
import numpy
from typing import Tuple
from aftab.functions import epsilon_greedy_vectorized
from .AftabBaseMixin import AftabBaseMixin


class AftabActionsMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    @torch.inference_mode()
    def get_action_tensors(
        self,
        q_values_train: torch.Tensor,
        q_values_test: torch.Tensor,
        epsilon_value: float,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if self._network.epsilon_greedy:
            actions_train = epsilon_greedy_vectorized(q_values_train, epsilon_value)
        else:
            actions_train = q_values_train.argmax(dim=-1)
        actions_test = q_values_test.argmax(dim=-1)
        return actions_train, actions_test

    @torch.inference_mode()
    def get_actions(
        self,
        q_values_train: torch.Tensor,
        q_values_test: torch.Tensor,
        epsilon_value: float,
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        actions_train, actions_test = self.get_action_tensors(
            q_values_train=q_values_train,
            q_values_test=q_values_test,
            epsilon_value=epsilon_value,
        )
        return actions_train.cpu().numpy(), actions_test.cpu().numpy()



import torch
from aftab.constants import ModuleType
from aftab.modules import LayerNorm2d

DEFAULT_ACTIVATION = torch.nn.ReLU


class EncoderBlock(torch.nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        kernel_size: int,
        stride: int,
        padding: int,
        activation: ModuleType = DEFAULT_ACTIVATION,
    ):
        super().__init__()
        self.convolutional = torch.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )
        self.activation = activation()
        self.normalization = LayerNorm2d(out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.convolutional(x)
        x = self.normalization(x)
        x = self.activation(x)
        return x



from aftab.networks import PQNNetwork
from aftab.networks import DuellingNetwork
from aftab.networks import BootstrappedNetwork
from aftab.networks import BootstrappedDuellingNetwork
from aftab.networks import DistributionalNetwork
from aftab.networks import DistributionalDuellingNetwork
from aftab.networks import DistributionalBootstrappedDuellingNetwork

networks_map = {
    "q": PQNNetwork,
    "duelling": DuellingNetwork,
    "bootstrapped": BootstrappedNetwork,
    "bootstrapped-duelling": BootstrappedDuellingNetwork,
    "distributional": DistributionalNetwork,
    "distributional-duelling": DistributionalDuellingNetwork,
    "distributional-bootstrapped-duelling": DistributionalBootstrappedDuellingNetwork,
    "bootstrapped-distributional-duelling": DistributionalBootstrappedDuellingNetwork,
}



import torch
from hl_gauss_pytorch import HLGaussLoss
from typing import Optional
from aftab.modules import Stream
from .BaseNetwork import BaseNetwork


class DistributionalBootstrappedDuellingNetwork(BaseNetwork):
    def __init__(
        self,
        *,
        distributional_bins: int,
        distributional_min_value: float,
        distributional_max_value: float,
        distributional_sigma: float,
        bootstrap_heads: int = 10,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if bootstrap_heads <= 0:
            raise ValueError("Expected `bootstrap_heads` to be positive.")

        self.distributional = True
        self.bootstrapped = True
        self.bootstrap_heads = bootstrap_heads
        self.distributional_bins = distributional_bins
        self.hl_gauss_loss = HLGaussLoss(
            min_value=distributional_min_value,
            max_value=distributional_max_value,
            num_bins=distributional_bins,
            sigma=distributional_sigma,
            clamp_to_range=True,
        )
        self.advantage_heads = torch.nn.ModuleList(
            [
                Stream(
                    input_dimension=self.feature_dimension,
                    hidden_dimension=self.embedding_dimension,
                    output_dimension=self.action_dimension * self.distributional_bins,
                    normalization=True,
                )
                for _ in range(self.bootstrap_heads)
            ]
        )
        self.value_heads = torch.nn.ModuleList(
            [
                Stream(
                    input_dimension=self.feature_dimension,
                    hidden_dimension=self.embedding_dimension,
                    output_dimension=self.distributional_bins,
                    normalization=True,
                )
                for _ in range(self.bootstrap_heads)
            ]
        )

    def get_value_logits_heads(self, features: torch.Tensor) -> torch.Tensor:
        values = [head(features) for head in self.value_heads]
        return torch.stack(values, dim=1).unsqueeze(2)

    def get_advantage_logits_heads(self, features: torch.Tensor) -> torch.Tensor:
        batch_size = features.size(0)
        advantages = [
            head(features).reshape(
                batch_size,
                self.action_dimension,
                self.distributional_bins,
            )
            for head in self.advantage_heads
        ]
        advantages = torch.stack(advantages, dim=1)
        return advantages - advantages.mean(dim=2, keepdim=True)

    def get_q_logits_heads(self, states: torch.Tensor) -> torch.Tensor:
        features = self.get_features(states)
        value_logits = self.get_value_logits_heads(features=features)
        advantage_logits = self.get_advantage_logits_heads(features=features)
        return value_logits + advantage_logits

    def gather_q_heads(
        self,
        q_heads: torch.Tensor,
        head_indices: torch.Tensor,
    ) -> torch.Tensor:
        action_dimension = q_heads.shape[-1]
        indices = head_indices.reshape(-1, 1, 1).expand(-1, 1, action_dimension)
        return q_heads.gather(1, indices).squeeze(1)

    def gather_q_logits_heads(
        self,
        q_logits_heads: torch.Tensor,
        head_indices: torch.Tensor,
    ) -> torch.Tensor:
        action_dimension = q_logits_heads.shape[-2]
        bins = q_logits_heads.shape[-1]
        indices = head_indices.reshape(-1, 1, 1, 1).expand(
            -1,
            1,
            action_dimension,
            bins,
        )
        return q_logits_heads.gather(1, indices).squeeze(1)

    def get_q_logits(
        self,
        states: torch.Tensor,
        head_indices: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        q_logits_heads = self.get_q_logits_heads(states)
        if head_indices is None:
            return q_logits_heads.mean(dim=1)
        return self.gather_q_logits_heads(
            q_logits_heads=q_logits_heads,
            head_indices=head_indices,
        )

    def get_q_heads(self, states: torch.Tensor) -> torch.Tensor:
        logits = self.get_q_logits_heads(states)
        q_values = self.hl_gauss_loss(
            logits.reshape(-1, self.distributional_bins),
        )
        return q_values.reshape(logits.shape[:-1])

    def get_q(
        self,
        states: torch.Tensor,
        head_indices: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        q_heads = self.get_q_heads(states)
        if head_indices is None:
            return q_heads.mean(dim=1)
        return self.gather_q_heads(q_heads=q_heads, head_indices=head_indices)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.get_q(x)



import torch


class LayerNorm2d(torch.nn.Module):
    def __init__(self, num_channels: int, epsilon: float = 1e-6):
        super().__init__()
        self.weight = torch.nn.Parameter(torch.ones(num_channels))
        self.bias = torch.nn.Parameter(torch.zeros(num_channels))
        self.epsilon = epsilon

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if x.dim() != 4:
            raise Exception("LayerNorm2d input must be four dimensional.")

        mean = x.mean(dim=1, keepdim=True)
        variance = (x - mean).pow(2).mean(dim=1, keepdim=True)
        x = (x - mean) / torch.sqrt(variance + self.epsilon)
        x = x * self.weight.view(1, -1, 1, 1) + self.bias.view(1, -1, 1, 1)
        return x



import torch
from aftab.constants import ModuleType
from aftab.common import LinearEpsilon


class BaseNetwork(torch.nn.Module):
    def __init__(
        self,
        *,
        action_dimension: int,
        embedding_dimension: int,
        encoder: ModuleType,
        channels_last: bool,
    ):
        super().__init__()
        self.channels_last = channels_last
        self.phi = encoder()

        self.epsilon_greedy = True
        self.epsilon = LinearEpsilon()

        self.action_dimension = action_dimension
        self.embedding_dimension = embedding_dimension
        dummy_input = self.__as_channels_last(torch.randn(1, 4, 84, 84))
        with torch.inference_mode():
            self.feature_dimension = self.phi(dummy_input).flatten(1).size(1)

    def __as_channels_last(self, x: torch.Tensor) -> torch.Tensor:
        if not self.channels_last or x.ndim != 4:
            return x
        return x.contiguous(memory_format=torch.channels_last)

    def no_epsilon_greedy(self) -> None:
        self.epsilon_greedy = False

    def normalize_observations(self, x: torch.Tensor) -> torch.Tensor:
        return x / 255.0

    def get_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.__as_channels_last(x)
        x = self.normalize_observations(x)
        x = self.__as_channels_last(x)
        features = self.phi(x)
        return features

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



import torch
from typing import Optional
from aftab.modules import Stream
from .BaseNetwork import BaseNetwork


class BootstrappedNetwork(BaseNetwork):
    def __init__(self, *, bootstrap_heads: int = 10, **kwargs):
        super().__init__(**kwargs)
        if bootstrap_heads <= 0:
            raise ValueError("Expected `bootstrap_heads` to be positive.")

        self.bootstrapped = True
        self.bootstrap_heads = bootstrap_heads
        self.q_heads = torch.nn.ModuleList(
            [
                Stream(
                    input_dimension=self.feature_dimension,
                    hidden_dimension=self.embedding_dimension,
                    output_dimension=self.action_dimension,
                )
                for _ in range(self.bootstrap_heads)
            ]
        )

    def get_q_heads(self, states: torch.Tensor) -> torch.Tensor:
        features = self.get_features(states)
        q_values = [head(features) for head in self.q_heads]
        return torch.stack(q_values, dim=1)

    def gather_q_heads(
        self,
        q_heads: torch.Tensor,
        head_indices: torch.Tensor,
    ) -> torch.Tensor:
        action_dimension = q_heads.shape[-1]
        indices = head_indices.reshape(-1, 1, 1).expand(-1, 1, action_dimension)
        return q_heads.gather(1, indices).squeeze(1)

    def get_q(
        self,
        states: torch.Tensor,
        head_indices: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        q_heads = self.get_q_heads(states)
        if head_indices is None:
            return q_heads.mean(dim=1)
        return self.gather_q_heads(q_heads=q_heads, head_indices=head_indices)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.get_q(x)



import torch
from aftab.modules import HadamaxBlock
from aftab.constants import ModuleType


class HadamaxZetaEncoder(torch.nn.Module):
    def __init__(self, *, activation: ModuleType = torch.nn.GELU):
        super().__init__()
        self.stream = torch.nn.Sequential(
            HadamaxBlock(
                4,
                48,
                kernel_size=4,
                stride=1,
                padding=2,
                pool_kernel=2,
                pool_stride=2,
                chi=activation,
                psi=activation,
            ),
            HadamaxBlock(
                48,
                48,
                kernel_size=4,
                stride=1,
                padding=2,
                pool_kernel=2,
                pool_stride=2,
                chi=activation,
                psi=activation,
            ),
            HadamaxBlock(
                48,
                48,
                kernel_size=4,
                stride=1,
                padding=2,
                pool_kernel=2,
                pool_stride=2,
                chi=activation,
                psi=activation,
            ),
            torch.nn.Flatten(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.stream(x)



from .NatureDQNEncoder import NatureDQNEncoder
from .AlphaEncoder import AlphaEncoder
from .BetaEncoder import BetaEncoder
from .GammaEncoder import GammaEncoder
from .DeltaEncoder import DeltaEncoder
from .EpsilonEncoder import EpsilonEncoder
from .ZetaEncoder import ZetaEncoder
from .EtaEncoder import EtaEncoder
from .ThetaEncoder import ThetaEncoder
from .HadamaxNatureDQNEncoder import HadamaxNatureDQNEncoder
from .HadamaxGammaEncoderV1 import HadamaxGammaEncoderV1
from .HadamaxGammaEncoderV2 import HadamaxGammaEncoderV2
from .HadamaxEpsilonEncoder import HadamaxEpsilonEncoder
from .HadamaxZetaEncoder import HadamaxZetaEncoder
from .HadamaxDeltaEncoder import HadamaxDeltaEncoder



import torch


class HadamaxLayerNorm2d(torch.nn.Module):
    def __init__(self, num_channels: int, epsilon: float = 1e-6):
        super().__init__()
        self.weight = torch.nn.Parameter(torch.ones(2 * num_channels))
        self.bias = torch.nn.Parameter(torch.zeros(2 * num_channels))
        self.epsilon = epsilon

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size, total_channels, height, width = x.shape
        C = total_channels // 2
        reshaped = x.view(batch_size, 2, C, height, width)
        variance, mean = torch.var_mean(reshaped, dim=2, keepdim=True, correction=0)
        normalized = (reshaped - mean) / torch.sqrt(variance + self.epsilon)
        weight = self.weight.view(1, 2, C, 1, 1)
        bias = self.bias.view(1, 2, C, 1, 1)
        normalized = normalized * weight + bias
        return normalized.view(batch_size, total_channels, height, width)



import torch
from aftab.modules import Stream
from .BaseNetwork import BaseNetwork


class DuellingNetwork(BaseNetwork):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.advantage = Stream(
            input_dimension=self.feature_dimension,
            hidden_dimension=self.embedding_dimension,
            output_dimension=self.action_dimension,
            normalization=True,
        )
        self.value = Stream(
            input_dimension=self.feature_dimension,
            hidden_dimension=self.embedding_dimension,
            output_dimension=1,
            normalization=True,
        )

    def get_value(self, features):
        return self.value(features)

    def get_advantage(self, features):
        advantage = self.advantage(features)
        advantage = advantage - advantage.mean(dim=1, keepdim=True)
        return advantage

    def get_q(self, states: torch.Tensor) -> torch.Tensor:
        features = self.get_features(states)
        value = self.get_value(features=features)
        advantage = self.get_advantage(features=features)
        return value + advantage

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.get_q(x)



from .Aftab import Aftab

from .constants import seeds
from .constants import seeds as SEEDS
from .constants import seeds as aftab_seeds

from .constants import environments
from .constants import environments as ENVS
from .constants import environments as aftab_environments

__version__ = "0.1.58"



from typing import Callable
from .BaseEncoder import BaseEncoder


def AlphaEncoder(activation: Callable):
    return BaseEncoder(
        configuration=[
            [4, 32, 4, 2, 1],
            [32, 64, 4, 2, 1],
            [64, 64, 3, 2, 1],
            [64, 64, 5, 0, 1],
        ],
        activation=activation,
    )



import torch
from typing import Type
from typing import Literal, Annotated

ModuleType = Type[torch.nn.Module]

EncoderStringType = Annotated[str, "must be a valid encoder key"]
OptimizerStringType = Literal["adam", "adamw" "radam", "nadam"]
NetworkStringType = Literal[
    "q",
    "duelling",
    "bootstrapped",
    "bootstrapped-duelling",
    "distributional",
    "distributional-duelling",
    "distributional-bootstrapped-duelling",
]


seeds = [475284, 219842, 525975, 909314]

environments = [
    "Alien-v5",
    "Amidar-v5",
    "Assault-v5",
    "Asterix-v5",
    "Asteroids-v5",
    "Atlantis-v5",
    "BankHeist-v5",
    "BattleZone-v5",
    "BeamRider-v5",
    "Berzerk-v5",
    "Bowling-v5",
    "Boxing-v5",
    "Breakout-v5",
    "Centipede-v5",
    "ChopperCommand-v5",
    "CrazyClimber-v5",
    "Defender-v5",
    "DemonAttack-v5",
    "DoubleDunk-v5",
    "Enduro-v5",
    "FishingDerby-v5",
    "Freeway-v5",
    "Frostbite-v5",
    "Gopher-v5",
    "Gravitar-v5",
    "Hero-v5",
    "IceHockey-v5",
    "Jamesbond-v5",
    "Kangaroo-v5",
    "Krull-v5",
    "KungFuMaster-v5",
    "MontezumaRevenge-v5",
    "MsPacman-v5",
    "NameThisGame-v5",
    "Phoenix-v5",
    "Pitfall-v5",
    "Pong-v5",
    "PrivateEye-v5",
    "Qbert-v5",
    "Riverraid-v5",
    "RoadRunner-v5",
    "Robotank-v5",
    "Seaquest-v5",
    "Skiing-v5",
    "Solaris-v5",
    "SpaceInvaders-v5",
    "StarGunner-v5",
    "Surround-v5",
    "Tennis-v5",
    "TimePilot-v5",
    "Tutankham-v5",
    "UpNDown-v5",
    "Venture-v5",
    "VideoPinball-v5",
    "WizardOfWor-v5",
    "YarsRevenge-v5",
    "Zaxxon-v5",
]



import torch
from .AftabBaseMixin import AftabBaseMixin


class AftabQValueMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    def __get_q_values_from_observations(self, observations: torch.Tensor):
        return self._network.get_q(observations)

    def get_q_values(
        self,
        float_train_observations: torch.Tensor = None,
        float_test_observations: torch.Tensor = None,
        float_observations: torch.Tensor = None,
        gradient: bool = False,
    ):
        with torch.set_grad_enabled(gradient):
            if float_observations is not None:
                return self.__get_q_values_from_observations(float_observations)
            train_count = float_train_observations.shape[0]
            q_values = self.__get_q_values_from_observations(
                torch.cat([float_train_observations, float_test_observations], dim=0)
            )
            return {"train": q_values[:train_count], "test": q_values[train_count:]}



import flax.linen as nn
from typing import Callable
from .BaseEncoder import BaseEncoder


def BetaEncoder(activation: Callable = nn.relu):
    return BaseEncoder(
        configuration=[
            [4, 32, 6, 2, 2],
            [32, 64, 3, 1, 1],
            [64, 32, 4, 2, 1],
            [32, 16, 8, 1, 0],
        ],
        activation=activation,
    )



import torch
from aftab.functions import lambda_returns
from .AftabBaseMixin import AftabBaseMixin


class AftabReturnsMixin(AftabBaseMixin):
    def __init__(self):
        super().__init__()

    @torch.inference_mode()
    def get_returns(
        self,
        *,
        batch_rewards: torch.Tensor,
        batch_terminations: torch.Tensor,
        next_q: torch.Tensor,
    ) -> torch.Tensor:
        return lambda_returns(
            rewards=batch_rewards,
            terminations=batch_terminations,
            next_q=next_q,
            gamma=getattr(self, "gamma"),
            return_lambda=getattr(self, "return_lambda"),
        )



import jax
import flax.linen as nn
from typing import Callable, List
from aftab.jax.modules import EncoderBlock


class BaseEncoder(nn.Module):
    activation: Callable
    configuration: List[List[int, int, int, int | str]]

    def setup(self):
        self.stream = [
            EncoderBlock(
                output_channels=output_channels,
                kernel=kernel,
                stride=stride,
                padding=padding,
            )
            for output_channels, kernel, stride, padding in self.configuration
        ]

    @nn.compact
    def __call__(self, x: jax.Array):
        batch_size = x.shape[0]
        for block in self.stream:
            x = block(x)
        return x.reshape(batch_size, -1)



import torch
from hl_gauss_pytorch import HLGaussLoss
from aftab.modules import Stream
from .BaseNetwork import BaseNetwork


class DistributionalNetwork(BaseNetwork):
    def __init__(
        self,
        *,
        distributional_bins: int,
        distributional_min_value: float,
        distributional_max_value: float,
        distributional_sigma: float,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.distributional = True
        self.distributional_bins = distributional_bins
        self.hl_gauss_loss = HLGaussLoss(
            min_value=distributional_min_value,
            max_value=distributional_max_value,
            num_bins=distributional_bins,
            sigma=distributional_sigma,
            clamp_to_range=True,
        )
        self.q_logits = Stream(
            input_dimension=self.feature_dimension,
            hidden_dimension=self.embedding_dimension,
            output_dimension=self.action_dimension * self.distributional_bins,
        )

    def get_q_logits(self, states: torch.Tensor) -> torch.Tensor:
        features = self.get_features(states)
        logits = self.q_logits(features)
        return logits.reshape(
            states.shape[0],
            self.action_dimension,
            self.distributional_bins,
        )

    def get_q(self, states: torch.Tensor) -> torch.Tensor:
        logits = self.get_q_logits(states)
        return self.hl_gauss_loss(logits)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.get_q(x)



import envpool
from .AftabBaseMixin import AftabBaseMixin


class AftabEnvironmentMixin(AftabBaseMixin):
    def _allocate_threads(self) -> tuple[int, int]:
        if self.cpu_count <= 1:
            return 1, 1
        total_envs = self.train_environments + self.test_environments
        train_ratio = self.train_environments / total_envs if total_envs > 0 else 0.8
        train_threads = max(1, int(self.cpu_count * train_ratio))
        test_threads = max(1, self.cpu_count - train_threads)
        return min(16, train_threads), min(16, test_threads)

    def make_environments(self, environment: str, seed: int):
        train_threads, test_threads = self._allocate_threads()

        common_kwargs = {
            "env_type": "gymnasium",
            "noop_max": self.noop,
            "frame_skip": self.frame_skip,
            "stack_num": self.frame_stack,
        }

        train_environment = envpool.make(
            environment,
            **common_kwargs,
            num_envs=self.train_environments,
            seed=seed,
            num_threads=train_threads,
            thread_affinity_offset=0,
            reward_clip=self.train_reward_clip,
            episodic_life=self.train_episodic_life,
        )

        test_environment = envpool.make(
            environment,
            **common_kwargs,
            num_envs=self.test_environments,
            seed=seed + self.seed_offset,
            num_threads=test_threads,
            thread_affinity_offset=train_threads if self.cpu_count > 1 else 0,
            reward_clip=self.test_reward_clip,
            episodic_life=self.test_episodic_life,
        )

        return (
            train_environment,
            test_environment,
            train_environment.action_space.n,
            train_environment.observation_space.shape,
        )



import torch
from aftab.modules import Stream
from .BaseNetwork import BaseNetwork


class PQNNetwork(BaseNetwork):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.q = Stream(
            input_dimension=self.feature_dimension,
            hidden_dimension=self.embedding_dimension,
            output_dimension=self.action_dimension,
        )

    def get_q(self, states: torch.Tensor) -> torch.Tensor:
        features = self.get_features(states)
        return self.q(features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.get_q(x)



import torch
from torch.nn.functional import cross_entropy
from torch.nn.functional import mse_loss
from typing import Optional


class AftabLossMixin:
    def __init__(self):
        super().__init__()

    def __get_distributional_logits_taken(
        self,
        *,
        q_logits: torch.Tensor,
        mini_batch_actions: torch.Tensor,
    ) -> torch.Tensor:
        action_indices = mini_batch_actions.reshape(-1, 1, 1).expand(
            -1,
            1,
            q_logits.shape[-1],
        )
        return q_logits.gather(1, action_indices).squeeze(1)

    def __get_distributional_cross_entropy(
        self,
        *,
        logits: torch.Tensor,
        target_probs: torch.Tensor,
    ) -> torch.Tensor:
        if logits.device.type == "cuda":
            with torch.autocast("cuda", enabled=False):
                return cross_entropy(logits, target_probs, reduction="none")
        return cross_entropy(logits, target_probs, reduction="none")

    def __get_distributional_loss(
        self,
        *,
        mini_batch_observations: torch.Tensor,
        mini_batch_actions: torch.Tensor,
        mini_batch_targets: torch.Tensor,
        mini_batch_old_q_values: Optional[torch.Tensor],
        mini_batch_target_probs: Optional[torch.Tensor],
    ) -> torch.Tensor:
        q_logits = self._network.get_q_logits(mini_batch_observations)
        q_logits_taken = self.__get_distributional_logits_taken(
            q_logits=q_logits,
            mini_batch_actions=mini_batch_actions,
        )
        hl_gauss = self._network.hl_gauss_loss
        if mini_batch_target_probs is None:
            loss = hl_gauss(q_logits_taken, mini_batch_targets, reduction="none")
        else:
            loss = self.__get_distributional_cross_entropy(
                logits=q_logits_taken,
                target_probs=mini_batch_target_probs,
            )

        value_clip = float(getattr(self, "distributional_value_clip", 0.0))
        if mini_batch_old_q_values is None or value_clip <= 0.0:
            return loss.mean()

        scalar_q_taken = hl_gauss(q_logits_taken)
        scalar_q_clipped = mini_batch_old_q_values + (
            scalar_q_taken - mini_batch_old_q_values
        ).clamp(-value_clip, value_clip)
        q_clipped_logprobs = hl_gauss.transform_to_logprobs(scalar_q_clipped)
        if mini_batch_target_probs is None:
            clipped_loss = hl_gauss(
                q_clipped_logprobs,
                mini_batch_targets,
                reduction="none",
            )
        else:
            clipped_loss = self.__get_distributional_cross_entropy(
                logits=q_clipped_logprobs,
                target_probs=mini_batch_target_probs,
            )
        return torch.max(loss, clipped_loss).mean()

    def get_loss(
        self,
        mini_batch_observations,
        mini_batch_actions,
        mini_batch_targets,
        mini_batch_old_q_values=None,
        mini_batch_target_probs=None,
    ) -> torch.Tensor:
        if bool(getattr(self._network, "distributional", False)):
            return self.__get_distributional_loss(
                mini_batch_observations=mini_batch_observations,
                mini_batch_actions=mini_batch_actions,
                mini_batch_targets=mini_batch_targets,
                mini_batch_old_q_values=mini_batch_old_q_values,
                mini_batch_target_probs=mini_batch_target_probs,
            )

        q_values = self._network.get_q(mini_batch_observations)
        q_taken = q_values.gather(1, mini_batch_actions.unsqueeze(1)).squeeze(1)
        return 0.5 * mse_loss(q_taken, mini_batch_targets)



import torch
from baloot import acceleration_device


def mse_loss(A, B):
    _device = acceleration_device()
    A = A.to(_device)
    B = B.to(_device)
    return 0.5 * torch.nn.functional.mse_loss(A, B)
